import stainless.lang._
import stainless.proof._

object Exercise5 {
  sealed abstract class Tree[T]
  case class Leaf[T]() extends Tree[T]
  case class Node[T](root: T, left: Tree[T], right: Tree[T]) extends Tree[T]

  def forall[T](t: Tree[T], p: T => Boolean): Boolean = {
    decreases(t)
    t match {
      case Leaf() => true
      case Node(root, left, right) => p(root) && forall(left, p) && forall(right, p)
    }
  }

  def isBalanced(t: Tree[BigInt]): Boolean = {
    decreases(t)
    t match {
      case Leaf() => true
      case Node(root, left, right) =>
        forall(left, (node: BigInt) => node <= root) &&
        forall(right, (node: BigInt) => node >= root) &&
        isBalanced(left) &&
        isBalanced(right)
    }
  }

  def insert(t: Tree[BigInt], x: BigInt): Tree[BigInt] = {
    require(isBalanced((t)))
    t match {
      case Leaf() => Node(x, Leaf(), Leaf())
      case Node(root, left, right) if (root > x) => Node(root, insert(left, x), right)
      case Node(root, left, right) if (x > root) => Node(root, left, insert(right, x))
      case _ => t // case where root == x
    }
  }

  def forallAfterInsert(t: Tree[BigInt], x: BigInt, p: BigInt => Boolean): Unit = {
    require(isBalanced(t) && forall(t, p) && p(x))
    t match {
      case Leaf() => ()
      case Node(root, left, right) if (root > x) => forallAfterInsert(left, x, p)
      case Node(root, left, right) if (x > root) => forallAfterInsert(right, x, p)
      case _ => ()
    }
  } ensuring(_ => forall(insert(t, x), p))

  def balancedAfterInsert(t: Tree[BigInt], x: BigInt): Unit = {
    require(isBalanced(t))
    t match {
      case Leaf() => ()
      case Node(root, left, right) if (root > x) =>
        balancedAfterInsert(left, x) // thanks to this recursive call, we know: isBalanced(insert(left, x)))
        forallAfterInsert(left, x, (node: BigInt) => node <= root)
        assert(insert(t , x) == Node(root, insert(left, x), right))
        assert(forall(insert(left, x), (node: BigInt) => node <= root))
        assert(forall(right, (node: BigInt) => node >= root))
        assert(isBalanced(left))
        assert(isBalanced(right))
        assert(isBalanced(insert(t, x)))
      case Node(root, left, right) if (x > root) =>
        balancedAfterInsert(right, x) // thanks to this recursive call, we know: isBalanced(insert(right, x)))
        forallAfterInsert(right, x, (node: BigInt) => node >= root)
        assert(isBalanced(insert(t, x)))
      case _ => ()
    }
  } ensuring(_ => isBalanced(insert(t, x)))
}
