import stainless.collection._
import stainless.annotation._

object Exercise6 {

  case class FSM[State, Input](states: List[State], transition: (State, Input) => List[State])

  def isRunFrom[State,Input](
    fsm: FSM[State, Input],
    run: List[(Input, State)],
    initial: State
  ): Boolean = {
    fsm.states.contains(initial) &&
    (run match {
      case Nil() => true
      case Cons((l,q), ps) =>
        fsm.transition(initial, l).contains(q) &&
        isRunFrom(fsm, ps, q)
    })
  }

  def isRunFromAllContained[State, Input](
    fsm: FSM[State, Input],
    run: List[(Input, State)],
    initial: State
  ): Unit = {
    require(isRunFrom(fsm, run, initial))

    if (!run.isEmpty) isRunFromAllContained(fsm, run.tail, run.head._2)

  } ensuring(_ =>
    fsm.states.contains(initial) &&
    run.map(_._2).forall(fsm.states.contains)
  )

  def isAcyclic[State,Input](run: List[(Input, State)], initial: State): Boolean = {
    val states = run.map(_._2)
    noDuplicate(Cons(initial, states))
  }

  def noDuplicate[T](l: List[T]): Boolean = l match {
    case Nil() => true
    case Cons(x, xs) => !xs.contains(x) && noDuplicate(xs)
  }

  def lengthRemoveNothing[T](@induct l: List[T], x: T): Unit = {
    require(!l.contains(x))

    ()

  } ensuring(_ => l.length == (l-x).length)

  def lengthRemoveOne[T](l: List[T], x: T): Unit = {
    require(l.contains(x))

    if (l.tail.contains(x)) {
      lengthRemoveOne(l.tail, x)
    } else {
      lengthRemoveNothing(l.tail, x)
    }

  } ensuring(_ => l.length > (l-x).length)

  def forallContained[T](l1: List[T], @induct l2: List[T], x: T): Unit = {
    require(l2.forall(l1.contains) && !l2.contains(x))

    ()
  } ensuring { _ =>
    val nl1 = l1 - x
    l2.forall(nl1.contains)
  }

  def largerLength[T](l1: List[T],l2: List[T]): Unit = {
    require(noDuplicate(l2) && l2.forall(l1.contains))

    if (!l2.isEmpty) {
      val nl1 = l1 - l2.head
      lengthRemoveOne(l1, l2.head)          // l1.length > nl1.length
      forallContained(l1, l2.tail, l2.head) // l2.tail.forall(nl1.contains)
      assert(noDuplicate(l2.tail))
      assert(l2.tail.forall(nl1.contains))
      largerLength(nl1, l2.tail)
    }

  } ensuring(_ => l1.length >= l2.length)

  def shortAcyclicRuns[State, Input](
    fsm: FSM[State, Input],
    run: List[(Input, State)],
    initial: State
  ): Unit = {
    require(isRunFrom(fsm, run, initial) && isAcyclic(run, initial))

    assert(noDuplicate(Cons(initial, run.map(_._2)))) // We know this because the run is acyclic

    isRunFromAllContained(fsm, run, initial) // this ensures that all states in the run and the initial states are in fsm.states
    assert(fsm.states.contains(initial))
    assert(run.map(_._2).forall(fsm.states.contains))

    largerLength(fsm.states, Cons(initial, run.map(_._2))) // fsm.states.length >= 1 + states.length

  } ensuring(_ => run.length < fsm.states.length)

}
